#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>

#include <sys/lkm.h>
#include <sys/device.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_ether.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/if_ether.h>
#endif

#include <dev/usb/usb.h>
#include <dev/usb/usb_port.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include <dev/usb/usbdevs.h>
#include "wimax.h"

#ifdef WIMAX_DEBUG
#define DPRINTF(x)  do { if (ehcidebug) printf x; } while(0)
#define DPRINTFN(n,x)   do { if (ehcidebug>(n)) printf x; } while (0)
int ehcidebug = 0;
#ifndef __NetBSD__
#define bitmask_snprintf(q,f,b,l) snprintf((b), (l), "%b", (q), (f))
#endif
#else
#define DPRINTF(x)
#define DPRINTFN(n,x)
#endif

CFDRIVER_DECL(wimax, DV_DULL, NULL);
USB_DECLARE_DRIVER(wimax);

void wimax_start(struct ifnet *);
int wimax_init(struct ifnet *);
void wimax_stop(struct ifnet *, int);
void wimax_watchdog(struct ifnet *);
int wimax_ioctl(struct ifnet *ifp, u_long command, void *data);

/* locators: port, config, iface, vendor, product, release */
static int usbdevloc[] = { -1, -1, -1, -1, -1, -1 }; 

static struct cfparent uhubparent = {
	"usbdevif", "uhub", DVUNIT_ANY
};

static struct cfdata wimax_cfdata[] = {
	{"wimax", "wimax", 0, FSTATE_STAR, usbdevloc, 0, &uhubparent},
	{ 0 }
};

static struct cfdriver *wimax_cfdrivers[] = {
	&wimax_cd,
	NULL
};
static struct cfattach *wimax_cfattachs[] = {
	&wimax_ca,
	NULL
};

static const struct cfattachlkminit wimax_cfattachinit[] = {
	{ "wimax", wimax_cfattachs },
	{ NULL }
};

static struct wimax_type wimax_devs[] = {
	{ 0x04e8, 0x6761 },
	{ 0x04e9, 0x6761 },
};


int wimax_lkmentry(struct lkm_table *, int, int);
MOD_DRV("wimax", wimax_cfdrivers, wimax_cfattachinit, wimax_cfdata);

static void fill_C_req(unsigned char *buf, int len)
{
	len -= 6;
	buf[0x00] = 0x57;
	buf[0x01] = 0x43;
	buf[0x02] = len & 0xff;
	buf[0x03] = (len >> 8) & 0xff;
	memset(buf + 6, 0, 12);
}

static int fill_normal_C_req(unsigned char *buf, unsigned short type_a, unsigned short type_b, unsigned short param_len, unsigned char *param)
{
	int len = 0x1a + param_len;
	fill_C_req(buf, len);
	buf[0x04] = 0x15;
	buf[0x05] = 0x00;
	buf[0x12] = 0x15;
	buf[0x13] = 0x00;
	buf[0x14] = type_a >> 8;
	buf[0x15] = type_a & 0xff;
	buf[0x16] = type_b >> 8;
	buf[0x17] = type_b & 0xff;
	buf[0x18] = param_len >> 8;
	buf[0x19] = param_len & 0xff;
	memcpy(buf + 0x1a, param, param_len);
	return len;
}

int fill_string_info_req(unsigned char *buf)
{
		return fill_normal_C_req(buf, 0x8, 0x1, 0x0, NULL);
}
int fill_init1_req(unsigned char *buf)
{
		unsigned char param[0x2] = {0x0, 0x1};
			return fill_normal_C_req(buf, 0x30, 0x1, sizeof(param), param);
}

int fill_mac_req(unsigned char *buf)
{
		return fill_normal_C_req(buf, 0x3, 0x1, 0x0, NULL);
}

int fill_init2_req(unsigned char *buf)
{
		return fill_normal_C_req(buf, 0x20, 0x8, 0x0, NULL);
}

int fill_init3_req(unsigned char *buf)
{
		return fill_normal_C_req(buf, 0x20, 0xc, 0x0, NULL);
}
int fill_authorization_data_req(unsigned char *buf)
{
		unsigned char param[0xd] = {0x00, 0x10, 0x00, 0x09, 0x40, 0x79, 0x6f, 0x74, 0x61, 0x2e, 0x72, 0x75, 0x00};
			return fill_normal_C_req(buf, 0x20, 0x20, sizeof(param), param);
}

int fill_outgoing_packet_header(unsigned char *buf, int len, int body_len)
{
	buf[0x00] = 0x57;
	buf[0x01] = 0x44;
	buf[0x02] = body_len & 0xff;
	buf[0x03] = (body_len >> 8) & 0xff;
	return body_len + 6;
}

inline void fill_config_req(unsigned char *buf, int body_len)
{
	fill_outgoing_packet_header(buf, 0x43, body_len);
	memset(buf + HEADER_LENGTH, 0, 12);
}

static int fill_normal_config_req(unsigned char *buf, unsigned short type_a, unsigned short type_b, unsigned short param_len, unsigned char *param) {
	int body_len = 0x14 + param_len;
	fill_config_req(buf, body_len);
	buf[0x04] = 0x15;
	buf[0x05] = 0x00;
	buf += HEADER_LENGTH;
	buf[0x0c] = 0x15;
	buf[0x0d] = 0x00;
	buf[0x0e] = type_a >> 8;
	buf[0x0f] = type_a & 0xff;
	buf[0x10] = type_b >> 8;
	buf[0x11] = type_b & 0xff;
	buf[0x12] = param_len >> 8;
	buf[0x13] = param_len & 0xff;
	memcpy(buf + 0x14, param, param_len);
	return HEADER_LENGTH + body_len;
}
	
int fill_init_cmd(unsigned char *buf)
{
	fill_config_req(buf, 0x12);
	buf[0x04] = 0x15;
	buf[0x05] = 0x04;
	buf += HEADER_LENGTH;
	buf[0x0c] = 0x15;
	buf[0x0d] = 0x04;
	buf[0x0e] = 0x50;
	buf[0x0f] = 0x04;
	return HEADER_LENGTH + 0x12;
}
		


int fill_diode_control_cmd(unsigned char *buf, int turn_on)
{
	unsigned char param[0x2] = {0x0, turn_on ? 0x1 : 0x0};
	return fill_normal_config_req(buf, 0x30, 0x1, sizeof(param), param);
}


int fill_protocol_info_req(unsigned char *buf, int len, unsigned char flags)
{
	buf[0x00] = 0x57;
	buf[0x01] = 0x45;
	buf[0x02] = 0x04;
	buf[0x03] = 0x00;
	buf[0x04] = 0x00;
	buf[0x05] = 0x02;
	buf[0x06] = 0x00;
	buf[0x07] = flags;
	return 8;
}

int fill_mac_lowlevel_req(unsigned char *buf)
{
	fill_outgoing_packet_header(buf, 0x50, 0x14);
	buf += HEADER_LENGTH_LOWLEVEL;
	memset(buf, 0, 0x14);
	buf[0x0c] = 0x15;
	buf[0x0d] = 0x0a;
	return HEADER_LENGTH_LOWLEVEL + 0x14;
}

USB_MATCH(wimax)
{
	int i;

	printf("-->wimax_match\n");

	USB_MATCH_START(wimax, uaa);

	printf("passed vid = %x, pid = %x\n", uaa->vendor, uaa->product);

	for (i = 0; i < __arraycount(wimax_devs); i++) {
		struct wimax_type *t = &wimax_devs[i];

		if (uaa->vendor == t->wimax_vid &&
				uaa->product == t->wimax_pid) {
			printf("Device match!\n");
			return(UMATCH_VENDOR_PRODUCT);
		}       
	}

	printf("<--wimax_match\n");

	return (UMATCH_NONE);
}

usbd_status
wimax_usb_read(struct wimax_softc *sc, u_int16_t length, u_int8_t *data)
{
	usbd_xfer_handle	xfer;
	usbd_status		err;
	int			total_len = 0, s;

	s = splnet();

	xfer = usbd_alloc_xfer(sc->wimax_udev);
	
	usbd_setup_xfer(xfer, sc->rx_pipe, 0, data, length, USBD_SHORT_XFER_OK,  1000, NULL); 

	err = usbd_sync_transfer(xfer);

	usbd_get_xfer_status(xfer, NULL, NULL, &total_len, NULL);

	printf("%s: transfered 0x%x bytes in\n",
				USBDEVNAME(sc->wimax_dev), total_len);

	usbd_free_xfer(xfer);

	splx(s);
	return(err);
}


usbd_status
wimax_usb_write(struct wimax_softc *sc, u_int16_t length, u_int8_t *data)
{
	usbd_xfer_handle	xfer;
	usbd_status		err;
	int			total_len = 0, s;

	s = splnet();

	xfer = usbd_alloc_xfer(sc->wimax_udev);
	
	usbd_setup_xfer(xfer, sc->tx_pipe, 0, data, length, USBD_SHORT_XFER_OK,  1000, NULL); 

	err = usbd_sync_transfer(xfer);

	usbd_get_xfer_status(xfer, NULL, NULL, &total_len, NULL);

	printf("%s: transfered 0x%x bytes out\n",
				USBDEVNAME(sc->wimax_dev), total_len);

	usbd_free_xfer(xfer);

	splx(s);
	return(err);
}

void recv_callback(usbd_xfer_handle xh, usbd_private_handle priv, usbd_status stat)
{
	char *buf;
	char count;
	printf("in callback!\n");
	usbd_get_xfer_status(xh, 0, &buf, &count, 0);
	printf("data: %s\n", buf);
}

usbd_status
wimax_async_read(struct wimax_softc *sc)
{
	usbd_xfer_handle	xfer;
	usbd_status		err;
	int			total_len = 0, s;

	s = splnet();

	xfer = usbd_alloc_xfer(sc->wimax_udev);
	
	usbd_setup_xfer(xfer, sc->rx_pipe, 0, sc->recv_buff, sc->bulk_in_len, USBD_SHORT_XFER_OK,  5000, recv_callback); 

	err = usbd_transfer(xfer);

	splx(s);
	return(err);

}


void wimax_start(struct ifnet *ifp)
{
	printf("--> wimax_start\n");

	printf("<-- wimax_start\n");
}

int wimax_init(struct ifnet *ifp)
{
	printf("--> wimax_init\n");

	printf("<-- wimax_init\n");
}

void wimax_stop(struct ifnet *ifp, int a) 
{
	printf("--> wimax_stop\n");

	printf("<-- wimax_stop\n");
}

void wimax_watchdog(struct ifnet *ifp)
{
	printf("--> wimax_watchdog\n");

	printf("<-- wimax_watchdog\n");
}

int wimax_ioctl(struct ifnet *ifp, u_long command, void *data)
{
	printf("--> wimax_ioctl\n");

	printf("<-- wimax_ioctl\n");
}

/*
int
wimax_send_command(struct wimax_softc *sc, u_int8_t *command, int size)
{
	return wimax_usb_write(sc, UT_WRITE_VENDOR_DEVICE, 0x0e, 0x0000,
			0x0000, size, command);
}
		    
int
wimax_get_cmd_status(struct wimax_softc *sc, u_int8_t cmd, u_int8_t *status)
{
	return wimax_usb_write(sc, UT_READ_VENDOR_INTERFACE, 0x22, cmd,
			0x0000, 6, status);
}
*/

/*
void analyze_ret(usbd_status st)
{
	switch (st) {
	case 	
}
*/

void wimax_get_config(struct wimax_softc *sc)
{
	unsigned char req_data[MAX_PACKET_LEN];
	int len;
	int r;
	int err;

	usbd_status st;

	char status[256];

	memset(status, 0, sizeof(status));

	unsigned char init_data1[] = {0x57, 0x45, 0x04, 0x00, 0x00, 0x02, 0x00, 0x74};

	unsigned char init_data2[] = {
		0x57, 0x50, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x15, 0x0a, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

	unsigned char init_data3[] = {
		0x57, 0x43, 0x12, 0x00, 0x15, 0x04, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x15, 0x04, 0x50, 0x04, 0x00, 0x00};

	printf("Trying to init ...\n");	

	st = wimax_async_read(sc);
	if (st != USBD_IN_PROGRESS) {
		printf("Cannot start async read: %s\n", usbd_errstr(st));
		return;
	}
	st = wimax_usb_write(sc, sizeof(init_data1), init_data1);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot initialize device(1): %s\n", usbd_errstr(st));
		return;
	}

	st = wimax_usb_write(sc, sizeof(init_data2), init_data2);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot initialize device(2): %s\n", usbd_errstr(st));
		return;
	}

	st = wimax_usb_write(sc, sizeof(init_data3), init_data3);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot initialize device(3): %s\n", usbd_errstr(st));
		return;
	}

	len=fill_string_info_req(req_data);

	st = wimax_usb_write(sc, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot send info request: %s\n", usbd_errstr(st));
		return;
	}

	len = fill_init1_req(req_data);
	st = wimax_usb_write(sc, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot send init1 request: %s\n", usbd_errstr(st));
		return;
	}

	len=fill_mac_req(req_data);
	st = wimax_usb_write(sc, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot send mac request: %s\n", usbd_errstr(st));
		return;
	}

	len=fill_string_info_req(req_data);
	st = wimax_usb_write(sc, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot send string info request: %s\n", usbd_errstr(st));
		return;
	}

	
	len=fill_init2_req(req_data);
	st = wimax_usb_write(sc, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot send info2 request: %s\n", usbd_errstr(st));
		return;
	}

	len=fill_init3_req(req_data);
	st = wimax_usb_write(sc, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot send info3 request: %s\n", usbd_errstr(st));
		return;
	}

	len=fill_authorization_data_req(req_data);
	st = wimax_usb_write(sc, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot send auth_data request: %s\n", usbd_errstr(st));
		return;
	}

	printf("Trying to control diode ...\n");	
	len = fill_diode_control_cmd(req_data, 0);
	st = wimax_usb_write(sc, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot send auth_data request: %s\n", usbd_errstr(st));
		return;
	}



	/*
	printf("Trying to request protocol info... ");	
	len = fill_protocol_info_req(req_data, MAX_PACKET_LEN,
			USB_HOST_SUPPORT_SELECTIVE_SUSPEND | USB_HOST_SUPPORT_DL_SIX_BYTES_HEADER |
			USB_HOST_SUPPORT_UL_SIX_BYTES_HEADER | USB_HOST_SUPPORT_DL_MULTI_PACKETS);

			
	st = wimax_usb_write(sc, UT_READ, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot complete usb request: %s\n", usbd_errstr(st));
		return;
	}
	printf("Trying to init ... ");	
	len = fill_init_cmd(req_data);
	st = wimax_usb_write(sc, UT_READ, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot complete usb request: %s\n", usbd_errstr(st));
	//	return;
	}

	*/
	/*
	printf("Trying to control diode ... ");	
	len = fill_diode_control_cmd(req_data, 1);
	st = wimax_usb_write(sc, UT_READ, len, req_data);
	*/

	//sleep(1);

	/*
	printf("Trying to control diode ... ");	
	len = fill_diode_control_cmd(req_data, 0);
	st = wimax_usb_write(sc, UT_WRITE, len, req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot complete usb request: %s\n", usbd_errstr(st));
	//	return;
	}
	//return;

	printf("Trying to request MAC(1) ... ");	
	st = wimax_usb_write(sc, UT_READ, sizeof(data2), data2);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot complete usb request: %s\n", usbd_errstr(st));
	//	return;
	}

	len = fill_mac_lowlevel_req(req_data);
	printf("\nTrying to request MAC(2) ... ");	
	st = wimax_usb_write(sc, UT_READ, sizeof(req_data), req_data);

	if (st != USBD_NORMAL_COMPLETION) {
		printf("Cannot complete usb request: %s\n", usbd_errstr(st));
	//	return;
	}

	//printf("usb request OK!\n");

	//set_data(req_data, len);
	//wimax_usb_write(&wimax_sc,
	//*/

	/*
	err = wimax_usb_write(sc, UT_WRITE, 0x33,
			0x0a02, 0x0000, sizeof(req_data),
			(u_int8_t *)&req_data);
			*/

	//	wimax_usb_write(sc, 
	//wimax_get_cmd_status(sc, 50, status);
}

void wimax_complete_attach(struct wimax_softc *sc)
{
	//struct ieee80211com     *ic = &sc->sc_ic;
	struct ifnet            *ifp = &sc->sc_if;
	usb_interface_descriptor_t  *id;
	usb_endpoint_descriptor_t   *ed;
	usbd_status         err;
	int             i;
	usbd_pipe_handle rx_pipe, tx_pipe;

	id = usbd_get_interface_descriptor(sc->wimax_iface);

	for (i = 0; i < id->bNumEndpoints; i++) {                                                                                                          
		ed = usbd_interface2endpoint_descriptor(sc->wimax_iface, i);                                                                                     
		if (!ed) {
			/*
			   printf(("%s: num_endp:%d\n", USBDEVNAME(sc->wimax_dev),                                                                                     sc->wimax_iface->idesc->bNumEndpoints)); */
			printf(("%s: couldn't get ep %d\n",
						(char *) USBDEVNAME(sc->wimax_dev), i));
			return;
		}

		if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_IN &&
				UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK) {
			sc->wimax_ed[WIMAX_ENDPT_RX] = ed->bEndpointAddress;
			sc->bulk_in_len = ed->bLength;
			printf("Found in endpoint: 0x%x\n", ed->bEndpointAddress);
		} else if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_OUT &&
				UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK) {
			sc->wimax_ed[WIMAX_ENDPT_TX] = ed->bEndpointAddress;
			printf("Found out endpoint: 0x%x\n", ed->bEndpointAddress);
		}

	}

//  usbd_status
//       usbd_open_pipe(usbd_interface_handle iface, u_int8_t address,
//       	 u_int8_t flags, usbd_pipe_handle *pipe);

	err = usbd_open_pipe(sc->wimax_iface, sc->wimax_ed[WIMAX_ENDPT_RX], NULL, &sc->rx_pipe);
	if (err) {
		printf("%s: tx fifo init failed\n", USBDEVNAME(sc->wimax_dev));
		return; 
	}

	err = usbd_open_pipe(sc->wimax_iface, sc->wimax_ed[WIMAX_ENDPT_TX], NULL, &sc->tx_pipe);
	if (err) {
		printf("%s: rx fifo init failed\n", USBDEVNAME(sc->wimax_dev));
		return; 
	}

	wimax_get_config(sc);

	ifp->if_softc = sc;
	memcpy(ifp->if_xname, USBDEVNAME(sc->wimax_dev), IFNAMSIZ);
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_init = wimax_init;
	ifp->if_stop = wimax_stop;
	ifp->if_start = wimax_start;
	ifp->if_ioctl = wimax_ioctl;
	ifp->if_watchdog = wimax_watchdog;
	ifp->if_mtu = WIMAX_DEFAULT_MTU;
	IFQ_SET_READY(&ifp->if_snd);

	if_attach(ifp);
}

USB_ATTACH(wimax)
{
	USB_ATTACH_START(wimax, sc, uaa);
	char                *devinfop;
	usbd_status         err;
	usbd_device_handle      dev = uaa->device;
	u_int8_t            mode, channel;
	int i;

	printf("-->wimax_attach\n");

	sc->wimax_dev = self;
	sc->sc_state = WIMAX_S_UNCONFIG;

	devinfop = usbd_devinfo_alloc(dev, 0);
	USB_ATTACH_SETUP;
	aprint_normal_dev(self, "%s\n", devinfop);
	usbd_devinfo_free(devinfop);
	
	err = usbd_set_config_no(dev, WIMAX_CONFIG_NO, 1);
	if (err) {
		aprint_error_dev(self, "setting config no failed\n");
		USB_ATTACH_ERROR_RETURN;
	}

	err = usbd_device2interface_handle(dev, WIMAX_IFACE_IDX, &sc->wimax_iface);
	if (err) {
		aprint_error_dev(self, "getting interface handle failed\n");
		USB_ATTACH_ERROR_RETURN;
	}

	sc->wimax_unit = device_unit(self);
	sc->wimax_udev = dev;

	wimax_complete_attach(sc);

	USB_ATTACH_SUCCESS_RETURN;
	///////////////
	//
	//
	//

}

USB_DETACH(wimax)
{
	USB_DETACH_START(wimax, sc);
	printf("-->wimax_detach\n");

	struct ifnet  *ifp = &sc->sc_if;
	if_detach(ifp);

	return (0);
}

int wimax_activate(device_ptr_t self, enum devact act)
{
	printf("-->wimax_activate\n");

	return (0);
}

int
wimax_lkmentry(struct lkm_table *lkmtp, int cmd, int ver)
{
	DISPATCH(lkmtp, cmd, ver, lkm_nofunc, lkm_nofunc, lkm_nofunc);
}

